#include <cstdio>
#include <iostream>
#include <string>
#include <vector>
#include <cstdarg>

/* FCM parser written by Bisqwit - http://iki.fi/bisqwit/ */

using namespace std;

static std::vector<unsigned char> data;
static unsigned R32(unsigned offs)
{
    // Note: This is not endianess-safe. Little-endian assumed!
    return *(unsigned *)&data[offs-4];
}
static unsigned R16(unsigned offs)
{
    // Note: This is not endianess-safe. Little-endian assumed!
    return *(unsigned short *)&data[offs-4];
}
/* sprintf() */
static const std::string format(const char* fmt, ...)
{
    char Buf[4096];
    using namespace std;
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(Buf, 4096, fmt, ap);
    va_end(ap);
    return Buf;
}
static string htmlspecialchars(const string& s)
{
    string result;
    for(unsigned a=0; a<s.size(); ++a)
    {
        if(s[a] == '"') { result += "&quot;";  continue; }
        if(s[a] == '&') { result += "&amp;";  continue; }
        if(s[a] == '<') { result += "&lt;";  continue; }
        result += s[a];
    }
    return result;
}

namespace
{
    static const char* const buttons[8]=
    {
        "A","B","Select","Start","Up","Down","Left","Right"
    };
    static const char* const commands[32] =
    {
        "Nop",
        "Reset",
        "Power cycle",
        "3","4","5","6",
        "VS coin", "VS dip 0",
        "9","10","11","12","13","14","15","16","17","18","19","20","21","22","23",
        "FDS ins","FDS eject","FDS side",
        "27","28","29","30","31"
    };
    
}

int main(int argc, const char* const * argv)
{
    FILE*fp = fopen(argv[1], "rb");
    if(!fp) return -1;
    fseek(fp, 0, SEEK_END);
    data.resize(ftell(fp));
    rewind(fp);
    fread(&data[0], 1, data.size(), fp);
    fclose(fp);
    
    unsigned rowlen=0;
    unsigned nframes=0;
    
    long saveoffs = -1;
    long ctrloffs = -1;
    
    enum { none, romname, movietitle } textmode = none;
    unsigned textbegin = 0;
    
    long ctrlmode = -1;
    long curframe = -1;
    string lastcmd = "";
    string text = "";
    
    unsigned frameoffs=0, frameshift=0;
    
    bool states[5][8] = {{false}};
    
    bool movieover = false;
    
    for(unsigned n=0; n<data.size(); )
    {
        text = "";
        if(!lastcmd.empty() && (ctrlmode == 0 || n+1 >= data.size()))
        {
            text = format("Frame %u: ", curframe) + lastcmd;
            
            if(n+1 < data.size())
            {
                cout << "  <span class=text>" << htmlspecialchars(text) << "</span>\n";
                rowlen = 0;
                lastcmd = "";
                text = "";
            }
        }
        
        if(ctrlmode >= 0 && curframe >= nframes)
        {
            if(!movieover) { movieover=true; cout << "<span class=movieover>"; }
        }
        
        if(!rowlen) cout << format("<span class=offs>%06X</span> ", n);

        unsigned char c = data[n++];
        
        cout << format("%02X ", c);
        
        if(n-1 == ctrloffs)
        {
            ctrlmode = 0; // begin ctrl handling.
            curframe = 0;
        }
        
        switch(ctrlmode)
        {
            case 0: // data
            {
                frameoffs  = 0;
                frameshift = 0;
                
                unsigned type = c>>7;
                ctrlmode = (c>>5)&3;
                unsigned cmd = (c&0x1F);
                
                if(type == 1)
                    lastcmd = format("Command: %s", commands[cmd]);
                else
                {
                    unsigned but = cmd & 7;
                    unsigned ctrl = cmd >> 3;
                  
                    states[ctrl][but] ^= 1;
                    
                    const char* s = states[ctrl][but] ? "PRESS" : "Release";
                    
                    lastcmd = format("P%u: %s %s", ctrl+1, s, buttons[but]);
                }
                break;
            }
            case 1:
            case 2:
            case 3:
                frameoffs |= c << frameshift;
                frameshift += 8;
                --ctrlmode;
                if(!ctrlmode)
                {
                    curframe += frameoffs;
                    // $text = "Offset +$frameoffs";
                }
                break;
            
            case -1:
                switch(textmode)
                {
                    case romname:
                        if(!c)
                        {
                            text = "rom name: \"" + string((char*)&data[textbegin], (char*)&data[n-1]) + "\"";
                            textmode = movietitle; textbegin = n;
                        }
                        break;
                    case movietitle:
                        if(!c)
                        {
                            text = "movie title: \"" + string((char*)&data[textbegin], (char*)&data[n-1]) + "\"";
                            textmode = none; textbegin = n;
                        }
                        break;
                    case none: ;
                }

                switch(n)
                {
                    case 0x00+4: text = "signature"; break;
                    case 0x04+4: text = "format version"; break;
                    case 0x08+4:
                    {
                        unsigned flags = R32(n);
                        const char* mode = (flags&4) ? "PAL" : "NTSC";
                        const char* begin = (flags&2) ? "reset" : "savestate";
                        text = string("flags: ") + mode + ", " + begin;
                        break;
                    }
                    case 0x0C+4:
                    {
                        nframes = R32(n);
                        text    = format("%u frames", nframes); break;
                    }
                    case 0x10+4: 
                    {
                        unsigned r = R32(n);
                        text = format("%u rerecords", r); break;
                    }
                    case 0x14+4:
                    {
                        unsigned ctrlbytes = R32(n);
                        text = format("%u bytes of controller data", ctrlbytes); break;
                    }
                    case 0x18+4:
                    {
                        saveoffs = R32(n);
                        text = format("savestate offset in file: %06X", saveoffs); break;
                    }
                    case 0x1C+4:
                    {
                        ctrloffs = R32(n);
                        text = format("controller data offset in file: %06X", ctrloffs); break;
                    }
                    case 0x20+16: text = "ROM md5sum"; break;
                    case 0x30+4:
                    {
                      unsigned ver = R32(n);
                      text = format("emulator version: 0.%.2f", (ver/100.0));
                      textmode = romname; textbegin = n;
                      break;
                    }
                }
                if(n==saveoffs)text = "end of metadata";
                if(n==ctrloffs)text = "end of savestate"; 
                if(n==data.size())text = "end of file";
                break; //ctrlmode -1
        }
        
        if(!text.empty())
        {
            cout << "  <span class=text>" << htmlspecialchars(text) <<  "</span>\n";
            rowlen = 0;
            if(n == saveoffs) cout << "<span class=state>";
            if(n == ctrloffs) cout << "</span>";
        }
        else
            if(++rowlen >= 128) { cout << "\n"; rowlen=0; }
    }
    if(movieover) cout << "</span>";
    cout << flush;
}
